#! /usr/local/bin/perl

# Script to generate the matrix-matrix mulitply code from the
# blocking combination input by the user.
# The routines generated are compatible with DGEMM or SGEMM.
# 
# Written by Dominic LAM <ctlam@cs.berkeley.edu>
# $Id: routine_gen.pl,v 1.1 1995/10/10 05:50:46 ctlam Exp $

$MM_GEN = "../mm_gen/mm_gen";               # name of mm generator
$PREC = "double";                 # default: dgemm
$matop = "NN";			  # defuult: NN

&parse_args;

# The names for the routines are below, but you shouldn't need to modify
# them.

if ($PREC eq "double"){
    $AB_b    = "mdmd_a1bc_md";
    $AB_ab   = "mdmd_acbc_md";
    $ABt_b   = "mdmdt_a1bc_md";
    $ABt_ab  = "mdmdt_acbc_md";
    $AtB_b   = "mdtmd_a1bc_md";
    $AtB_ab  = "mdtmd_acbc_md";
    $AtBt_b  = "mdtmdt_a1bc_md";
    $AtBt_ab = "mdtmdt_acbc_md";
} else {
    $AB_b    = "mfmf_a1bc_mf";
    $AB_ab   = "mfmf_acbc_mf";
    $ABt_b   = "mfmft_a1bc_mf";
    $ABt_ab  = "mfmft_acbc_mf";
    $AtB_b   = "mftmf_a1bc_mf";
    $AtB_ab  = "mftmf_acbc_mf";
    $AtBt_b  = "mftmft_a1bc_mf";
    $AtBt_ab = "mftmft_acbc_mf";
}    


if (!-e $MM_GEN){		# make the generator if it's not exist
    print "Making $MM_GEN......\n";
    `gcc -O -o $MM_GEN $MM_GEN.c`;
    die ("No generator available.\n") if ($?);
}

print "Generating matrix-matrix code......\n";
&gen_code;
`make -f makefile.dgemm clean`;

exit(0);


sub gen_code {
    
    local(@cmd, $i);

    @cmd = ("./".$MM_GEN);
    
    for ($i=0; $i<@m; $i++){
	push (@cmd, "-cb");
	push (@cmd, $m[$i]);
	push (@cmd, $k[$i]);
	push (@cmd, $n[$i]);
    }
    if ($PREC eq "double"){
	push (@cmd, "-prec double");
    } else {
	push (@cmd, "-prec single");
    }

    if ($matop eq "NN"){
    
	# C = AB + beta*C  
	`@cmd -file $AB_b.c -routine_name $AB_b -beta c`;
	die ("Problem in generating C = AB + beta*C code\n") if ($?);

	# C = alpha*AB + beta*C
	`@cmd -file $AB_ab.c -routine_name $AB_ab -alpha c -beta c`;
	die ("Problem in generating C = alpha*AB + beta*C code\n") if ($?);

	# Switching parameters and generate TT as well 

	@cmd = ();
	@cmd = ("./".$MM_GEN);
	for ($i=0; $i<@m; $i++){
	    push(@cmd, "-cb");
	    push(@cmd, $n[$i]);
	    push(@cmd, $k[$i]);
	    push(@cmd, $m[$i]);
	}
	if ($PREC eq "double"){
	    push (@cmd, "-prec double");
	} else {
	    push (@cmd, "-prec single");
	}

	# C = A'B' + beta*C
	`@cmd -file $AtBt_b.c -routine_name $AtBt_b -beta c -opA T -opB T`;
	die ("Problem in generating C = A'B' + beta*C code\n") if ($?);
	
	# C = alpha*A'B' + beta*C
	`@cmd -file $AtBt_ab.c -routine_name $AtBt_ab -alpha c -beta c -opA T -opB T`;
	die ("Problem in generating C = alpha*A'B' + beta*C code\n") if ($?);
    }

    if ($matop eq "NT"){

	# C = AB' + beta*C
	`@cmd -file $ABt_b.c -routine_name $ABt_b -beta c -opB T`;
	die ("Problem in generating C = AB' + beta*C code\n") if ($?);

	# C = alpha*AB' + beta*C 
	`@cmd -file $ABt_ab.c -routine_name $ABt_ab -alpha c -beta c -opB T`;
	die ("Problem in generating C = alpha*AB' + beta*C code\n") if ($?);
    }

    if ($matop eq "TN"){

	# C = A'B + beta*C
	`@cmd -file $AtB_b.c -routine_name $AtB_b -beta c -opA T`;
	die ("Problem in generating C = A'B + beta*C code\n") if ($?);

	# C = alpha*A'B + beta*C
	`@cmd -file $AtB_ab.c -routine_name $AtB_ab -alpha c -beta c -opA T`;
	die ("Problem in generating C = alpha*A'B + beta* code\n") if ($?);
    }

    if ($matop eq "TT"){

	# C = A'B' + beta*C
	`@cmd -file $AtBt_b.c -routine_name $AtBt_b -beta c -opA T -opB T`;
	die ("Problem in generating C = A'B' + beta*C code\n") if ($?);
	
	# C = alpha*A'B' + beta*C
	`@cmd -file $AtBt_ab.c -routine_name $AtBt_ab -alpha c -beta c -opA T -opB T`;
	die ("Problem in generating C = alpha*A'B' + beta*C code\n") if ($?);
    }

}
	
sub parse_args {

    local ($i);
    local ($m, $k, $n);

    if (!@ARGV){
	&help_menu;
	exit(1);
    }

    for ($i=0; $i<@ARGV; $i++){
	if ($ARGV[$i] =~ /^-./){

	    # Test for the -cb option
	    if ($ARGV[$i] =~ /^-cb\b/){
		$m = $ARGV[++$i];
		$k = $ARGV[++$i];
		$n = $ARGV[++$i];
		die ("M must be an integer.\n") if ($m =~ /\D+/);
		die ("K must be an integer.\n") if ($k =~ /\D+/);
		die ("N must be an integer.\n") if ($n =~ /\D+/);
		push(@m, $m);
		push(@k, $k);
		push(@n, $n);
	    }
	    
	    # Test for the precision option
	    elsif ($ARGV[$i] =~ /^-prec\b/){
		$PREC = $ARGV[++$i];
		die ("Must be either 'double' or 'single'.\n")
		    if ($PREC ne "single" && $PREC ne "double");
	    }

	    # Test for matop option
	    elsif ($ARGV[$i] =~ /^-matop\b/){
		$matop = $ARGV[++$i];
		die ("matop must be NN|NT|TN|TT\n")
		    if ($matop ne "NN" && 
			$matop ne "NT" &&
			$matop ne "TN" &&
			$matop ne "TT");
	    }
		
	    # Test for help option
	    elsif ($ARGV[$i] =~ /^-help\b/){
		&help_menu;
		exit(0);
	    }

	    # defaults
	    else {
		print "Unknown option $ARGV[$i], ignored.\n";
	    }
	} else {
	    print "Unknown argument $ARGV[$i], ignored.\n";
	}
    }

}

sub help_menu {

    print "Usage : $0 [OPTIONS]\n";
    print "where [OPTIONS] include:\n";
    print " -help \t\t\t\t: Print this message\n";
    print " -cb m k n [-cb m k n]\t\t: blocking sizes\n";
    print " -prec <double>|single]\t\t: precision used\n";
    print " -matop NN|NT|TN|TT\t\t: matrix operation\n";

}
