/*
    $Id: IPM_test.c,v 1.14 1995/09/12 01:23:43 krste Exp $

    Test wrapper for IPM routines.
*/

#include <stdio.h>

#include "IPM_timer.h"

static char foo[100];

/* Timing loop parameters. */
#define CLEAR_LOOPS (2)
#define ACC_LOOPS (10)
#define INNER_LOOPS (10000)

int
main()
{
    IPM_timer t;
    long k;
    long j;
    long i;
    long sum = 0;
    volatile char* p = foo;
    char buf[IPM_TIMER_REPORT_LEN+20];

    /* Check out range and resolution calls. */
    printf("Timer range %3g s, resolution %4g s\n",
           IPM_timer_range(),
           IPM_timer_resolution());

    /* Try interval timing and timer report. */
    for (k=0; k<CLEAR_LOOPS; k++)
    {
        IPM_timer_clear(&t);

        for (j=0; j<ACC_LOOPS; j++)
        {
            IPM_timer_start(&t);

            for (i=0; i<INNER_LOOPS; i++)
            {
                sum += p[(i % 12) + 2];
            }

            IPM_timer_stop(&t);

            IPM_timer_report("testing", &t, buf);

            printf("%2ld: Timer reports\n%s", j, buf);
        }
        printf("\n");
    }

    /* Try out IPM_timer_read call. */
    for (k=0; k<CLEAR_LOOPS; k++)
    {
        IPM_timer_clear(&t);

        for (j=0; j<ACC_LOOPS; j++)
        {
            IPM_timer_start(&t);

            for (i=0; i<INNER_LOOPS; i++)
            {
                sum += p[(i % 12) + 2];
            }

            IPM_timer_stop(&t);

            printf("%2ld: IPM_timer_read = %g s.\n", j, IPM_timer_read(&t));
        }
        printf("\n");
    }

    /* Try out IPM_timer_get_time call. */

    for (k=0; k<CLEAR_LOOPS; k++)
    {
        double t;
        double last_t = 0;

        for (j=0; j<ACC_LOOPS; j++)
        {
            t = IPM_timer_get_time();
            printf("%2ld: IPM_timer_get_time = %20.18g s, diff = %20.18g s.\n",
                   j,
                   t,
                   t - last_t);

            last_t = t;

            for (i=0; i<INNER_LOOPS; i++)
            {
                sum += p[(i % 12) + 2];
            }
        }

        printf("\n");
    }

    return 0;
}
