const char* IPM_timer_gettimer_rcsid = "$Header: /u/spert/src/IPM/RCS/IPM_timer_gettimer.c,v 1.10 1995/09/12 01:17:55 krste Exp $";

/******************************************************/
/* Code that uses gettimer() to implement the timer.  */
/******************************************************/

#include <assert.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>

#include "IPM_timer.h"
#include "IPM_timer_gettimer.h"

typedef struct
{
    struct timestruc_t start;   /* Start values. */
    struct timestruc_t stop;    /* Stop values. */
    double n_intervals;         /* Number of intervals measured. */
    double t_ns;                /* Accumulated nanoseconds, stored in */
                                /* double. */
} IPM_timer_gettimer;

static void
IPM_gettimer_abort(const char *message)
{
    perror(message);
    fflush(stderr);
    abort();
}

/* Cache resolution and range info. */
static double resolution;
static double range;

static double
IPM_timer_resolution_gettimer(void)
{ return resolution; }

static double
IPM_timer_range_gettimer(void)
{ return range; }

static void
IPM_timer_clear_gettimer(IPM_timer *ipmt)
{
    IPM_timer_gettimer *gt = (IPM_timer_gettimer *) ipmt;

    gt->n_intervals = 0.0;       /* Clear number of intervals. */
    gt->t_ns = 0.0;              /* Clear accumulated time. */
}

static void
IPM_timer_start_gettimer(IPM_timer *ipmt)
{
    IPM_timer_gettimer *gt = (IPM_timer_gettimer *) ipmt;

    if (gettimer(TIMEOFDAY, &(gt->start)))
        IPM_gettimer_abort("IPM_timer_start_gettimer()");
}

/* Do timestruc_t calculations in nanoseconds as this avoids the accuracy */
/* loss that would occur when converting 1e-9 to binary if working in    */
/* seconds. */
#define TS2NS(ts) \
    (((ts).tv_sec * 1e9) + (ts).tv_nsec)

static void
IPM_timer_stop_gettimer(IPM_timer *ipmt)
{
    IPM_timer_gettimer *gt = (IPM_timer_gettimer *) ipmt;

    if (gettimer(TIMEOFDAY, &(gt->stop)))
        IPM_gettimer_abort("IPM_timer_stop_gettimer()");
    gt->n_intervals += 1.0;
    gt->t_ns += TS2NS(gt->stop) - TS2NS(gt->start);
}

static double
IPM_timer_read_gettimer(const IPM_timer *ipmt)
{
    const IPM_timer_gettimer *gt = (const IPM_timer_gettimer *) ipmt;
    
    /* TODO: Subtract off calling overhead. */
    return gt->t_ns * 1e-9;
}

static void
IPM_timer_report_gettimer(const char *tag,
                 const IPM_timer *ipmt,
                 char *buf)
{
    /* TODO: Format string makes assumptions about resolution. */
    sprintf(buf, "INTERVAL: %s\nREAL_TIME: %10.7f\n",
            tag,
            IPM_timer_read_gettimer(ipmt));
}

static double
IPM_timer_get_time_gettimer(void)
{
    struct timestruc_t t;
    
    if (gettimer(TIMEOFDAY, &t))
        IPM_gettimer_abort("IPM_timer_get_time_gettimer()");

    return TS2NS(t) * 1e-9;
}

void
IPM_init_gettimer(void)
{
    struct timestruc_t res;
    struct timestruc_t max_value;

    if (restimer(TIMEOFDAY, &res, &max_value))
        IPM_gettimer_abort("IPM_init_gettimer()");

    resolution = TS2NS(res) * 1e-9;
    range = TS2NS(max_value) * 1e-9;

    /* Make sure opaque struct large enough. */
    assert(sizeof(IPM_timer_gettimer) <= IPM_TIMER_STRUCT_MAX_SIZE);

    /* TODO: Get calling overhead from arguments. */

    /* Install function pointers. */
    IPM_timer_resolution_p = IPM_timer_resolution_gettimer;
    IPM_timer_range_p = IPM_timer_range_gettimer;
    IPM_timer_clear_p = IPM_timer_clear_gettimer;
    IPM_timer_start_p = IPM_timer_start_gettimer;
    IPM_timer_stop_p = IPM_timer_stop_gettimer;
    IPM_timer_read_p = IPM_timer_read_gettimer;
    IPM_timer_report_p = IPM_timer_report_gettimer;
    IPM_timer_get_time_p = IPM_timer_get_time_gettimer;
}
