const char* IPM_timer_syssgi_rcsid = "$Header: /u/spert/src/IPM/RCS/IPM_timer_syssgi.c,v 1.11 1995/09/12 01:00:11 krste Exp $";

/**************************************************************/
/* Code to use the hardware real time counter on SGI systems. */
/**************************************************************/

#include "IPM_timer.h"
#include "IPM_timer_syssgi.h"

#include <assert.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/syssgi.h>
#include <sys/mman.h>
#include <stdio.h>

typedef struct
{
    unsigned long start;
    unsigned long n_intervals;
    unsigned long total;
} IPM_timer_syssgi;

extern int getpagesize(void);

static volatile unsigned long* clock_p;

static double resolution;       /* Resolution in seconds. */

static double
IPM_timer_resolution_syssgi(void)
{ return resolution; }

static double
IPM_timer_range_syssgi(void)
{
    /* Guess 32-bit counter. */
    return (double) 0xffffffffUL * resolution;
}

static void
IPM_timer_clear_syssgi(IPM_timer *ipmt)
{
    IPM_timer_syssgi *sst = (IPM_timer_syssgi *) ipmt;
    
    sst->n_intervals = 0UL;
    sst->total = 0UL;
}

static void
IPM_timer_start_syssgi(IPM_timer *ipmt)
{
    IPM_timer_syssgi *sst = (IPM_timer_syssgi *) ipmt;

    sst->start = *clock_p;
}

static void
IPM_timer_stop_syssgi(IPM_timer *ipmt)
{
    IPM_timer_syssgi *sst = (IPM_timer_syssgi *) ipmt;

    const unsigned long stop = *clock_p;

    sst->n_intervals++;

    sst->total += (stop - sst->start);
}

static double
IPM_timer_read_syssgi(const IPM_timer *ipmt)
{
    const IPM_timer_syssgi *sst = (const IPM_timer_syssgi *) ipmt;

    /* TODO: Subtract calling overhead. */
    return sst->total * resolution;
}

static void
IPM_timer_report_syssgi(const char* tag,
                        const IPM_timer *ipmt,
                        char* buf)
{
    sprintf(buf, "INTERVAL: %s\nREAL_TIME: %4.8f\n",
            tag,
            IPM_timer_read_syssgi(ipmt));
}

static double
IPM_timer_get_time_syssgi(void)
{
    return (*clock_p) * resolution;
}

static void
IPM_init_counter(void)
{
    int rval;                   /* Return value from syssgi(). */
    int mem_fd;                 /* File descriptor of system memory. */
    int pagesize;               /* Size of system pages. */
    int page_addr;              /* Page aligned address of counter. */
    int clock_addr_lo;          /* Low bits of clock address. */
    void *paddr;                /* Virtual page address returned by mmap(). */
    int clock_addr;             /* Virtual clock address on page. */
    unsigned int picoseconds;   /* Resolution in picoseconds. */

    /* Check for cycle counter. */
    rval = syssgi(SGI_QUERY_CYCLECNTR, &picoseconds);
    if (rval == -1)
    {
        fprintf(stderr,
                "IPM_init_syssgi(): No hardware cycle counter present.\n");
        fflush(stderr);
        abort();
    };
    /* fprintf(stderr, "rval=0x%08x\n", rval); */

    mem_fd = open("/dev/mmem", O_RDONLY);
    if (mem_fd == -1)
    {
        perror("IPM_init():open()");
        abort();
    }
    
    /* Map cycle counter into page-aligned user memory. */
    pagesize = getpagesize();
    clock_addr_lo = rval & (pagesize - 1);
    page_addr = rval - clock_addr_lo;
    paddr = mmap(0,                  /* Don't care where in virtual space. */
                 pagesize,           /* Only need one page. */
                 PROT_READ,          /* Only need to read. */
                 MAP_SHARED,         /* Simplest option. */
                 mem_fd,             /* From /dev/mem open command. */
                 (off_t) page_addr); /* Clock's physical address. */

    if ((int) paddr == -1)
    {
        perror("IPM_init():mmap():");
        abort();
    }
    /* fprintf(stderr, "paddr=0x%08x\n", (unsigned int) paddr); */

    clock_addr = ((int) paddr) + clock_addr_lo;
    clock_p = (unsigned long *) clock_addr;

    resolution = picoseconds * 1e-12;  /* Convert picoseconds to seconds. */
}

void
IPM_init_syssgi(void)
{
    /* Make sure opaque struct large enough. */
    assert(sizeof(IPM_timer_syssgi) <= IPM_TIMER_STRUCT_MAX_SIZE);

    /* Set up counter. */
    IPM_init_counter();

    /* Install function pointers. */
    IPM_timer_resolution_p = IPM_timer_resolution_syssgi;
    IPM_timer_range_p = IPM_timer_range_syssgi;
    IPM_timer_clear_p = IPM_timer_clear_syssgi;
    IPM_timer_start_p = IPM_timer_start_syssgi;
    IPM_timer_stop_p = IPM_timer_stop_syssgi;
    IPM_timer_read_p = IPM_timer_read_syssgi;
    IPM_timer_report_p = IPM_timer_report_syssgi;
    IPM_timer_get_time_p = IPM_timer_get_time_syssgi;
}
