const char* IPM_timer_t0vcount_rcsid = "$Header: /u/spert/src/IPM/RCS/IPM_timer_t0vcount.c,v 1.8 1995/09/12 03:28:08 krste Exp $";

/***********************************************/
/* Code to use the T0 counter on a SPERT board */
/***********************************************/

#include <assert.h>
#include <stdio.h>

#include "IPM_timer.h"
#include "IPM_timer_t0vcount.h"

/******************************************************************/
/* These parameters should really be retreived via a system call. */
/******************************************************************/

/* Assume 40 MHz boards for now (25 ns cycle time). */
#ifndef IPM_T0VCOUNT_RESOLUTION
#  define IPM_T0VCOUNT_RESOLUTION (25e-9)
#endif

typedef struct
{
    unsigned long start;
    unsigned long n_intervals;
    unsigned long total;
} IPM_timer_t0vcount;

static double resolution;	/* Resolution of the clock in seconds */
static double range;            /* Range of the clock in seconds. */

/*
    This routine reads the vcount register from coprocessor 2.  This
    holds a counter that is incremented once per clock cycle.
*/

#ifdef __GNUC__
static unsigned long
t0clock(void)
{
    register unsigned long res;

    __asm__ volatile ("cfvu %0, $1" : "=r" (res) : );
    return res;
}
#else
#error "Can only be compiled with gcc"
#endif

static double
IPM_timer_resolution_t0vcount(void)
{ return resolution; }

static double
IPM_timer_range_t0vcount(void)
{ return range; }

static void
IPM_timer_clear_t0vcount(IPM_timer *ipmt)
{
    IPM_timer_t0vcount *s0vc = (IPM_timer_t0vcount *) ipmt;

    s0vc->n_intervals = 0UL;
    s0vc->total = 0UL;
}

static void
IPM_timer_start_t0vcount(IPM_timer *ipmt)
{
    IPM_timer_t0vcount *s0vc = (IPM_timer_t0vcount *) ipmt;

    s0vc->start = t0clock();
}

static void
IPM_timer_stop_t0vcount(IPM_timer *ipmt)
{
    IPM_timer_t0vcount *s0vc = (IPM_timer_t0vcount *) ipmt;

    const unsigned long stop = t0clock(); /* No need to store this. */

    s0vc->n_intervals++;
    s0vc->total += (stop - s0vc->start);
}

static double
IPM_timer_read_t0vcount(const IPM_timer *ipmt)
{
    const IPM_timer_t0vcount *s0vc = (const IPM_timer_t0vcount *) ipmt;

    /* TODO: Subtract calling overhead. */
    return (double) s0vc->total * resolution;
}

static void
IPM_timer_report_t0vcount(const char* tag,
                              const IPM_timer *ipmt,
                              char* buf)
{
    sprintf(buf, "INTERVAL: %s\nREAL_TIME: %4.8f\n",
            tag,
            IPM_timer_read_t0vcount(ipmt));
}

static double
IPM_timer_get_time_t0vcount()
{
    return (double) t0clock() * resolution;
}

void
IPM_init_t0vcount(void)
{
    /* TODO - we should get the clock frequency here to calculate
       the resolution - SPERT boards have variable speed clocks. */
    resolution = IPM_T0VCOUNT_RESOLUTION;

    /* 32-bit counter. */
    range = IPM_T0VCOUNT_RESOLUTION * (double) 0xffffffffUL;

    /* Make sure opaque struct large enough. */
    assert(sizeof(IPM_timer_t0vcount) <= IPM_TIMER_STRUCT_MAX_SIZE);

    /* Install function pointers. */
    IPM_timer_resolution_p = IPM_timer_resolution_t0vcount;
    IPM_timer_range_p = IPM_timer_range_t0vcount;
    IPM_timer_clear_p = IPM_timer_clear_t0vcount;
    IPM_timer_start_p = IPM_timer_start_t0vcount;
    IPM_timer_stop_p = IPM_timer_stop_t0vcount;
    IPM_timer_read_p = IPM_timer_read_t0vcount;
    IPM_timer_report_p = IPM_timer_report_t0vcount;
    IPM_timer_get_time_p = IPM_timer_get_time_t0vcount;
}
