#! /usr/local/bin/perl

# This program checks for the optimal blocking for the matrix-matrix
# multiplication code C = A*B + C
# 
# $Id: mm_search.pl,v 1.9 1995/10/14 21:47:55 ctlam Exp $
# 
# Written by: Dominic Lam <ctlam@cs.berkeley.edu>
# 
# "Copyright (c) 1995 The Regents of the University of California.  
# All rights reserved."
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without written agreement is
# hereby granted, provided that the above copyright notice and the following
# two paragraphs appear in all copies of this software.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#  
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
# ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
# 
 
## COMPILE FLAGS ###
                                # flags for compilation
$CCFLAGS = "-I../rprf -I../IPM";
$CCLIBS = "-lm ../IPM/libipm.a";
$CCLIBS = $CCLIBS ." ../rprf/librprf.a";
$CCDEF = "";			# Nothing for this version


# ------------ You don't want to modify anything below ---------------- #

### FILES ###
$GEN_TARGET = "../mm_gen/mm_gen";
$FLNAME = "mm";
$PERFNAME = "./mul_mm_m_perf";
$TARGET = "./mm_perf";		# compile target
$RES_FILE = "./mm_res_"."L0_NN.mat"; 
$outfile = "./mm.result";          # result file for a specific architecture
$GUESS = "../IPM/config.guess";	   # architecture guesser
$LOG_FILE = "./mm_search.$$.log";  # log file

### FLAGS ###
$CBFLAG = "-cb";		# flags for mm-generation
$FLFLAG = "-file";
$ROUTINE_NAME = "php_mul_mfmf_mf";
$ROUTINEFLAG = "-routine_name";
$PRECFLAG = "-prec";
$OPA = "-opA";
$OPB = "-opB";


### DEFAULTS ###                # may be changed by user
                                # default compiler : acc
$COMPILER = "cc";
$depth = 1;                     # default memory hierarchy
$nbest = 10;                    # number of best block combination
$reg_num = 32;			# no. of register in the machine (32)
$k_limit = 4;			# limit of k for trials (4)
$k_allow = 5;                   # number of consective decrease before
                                # being cut
$L1 = 8;                       # L1 cache size
$L2 = 0;                        # L2 cache size
$max_mat1 = 0;                  # default maximum matrix size for L1
$max_mat2 = 0;			# default maximum matrix size for L2
$ver = 0;                       # output file level number
$next_ver = 0;			# next version
$var = 0.01;                    # acceptable limit for perk
$sizecutoff = 300;		# cut off size of multiples of 10
$start_at = 0;			# level of search starts
                                # allowance for peak variation
@PEAK_ALLOW = (0.25, 0.5, 1.0, 1.5, 2.0);
@PEAK_ALLOWN = (0.5, 1.0, 2.0, 3.0, 4.0);
$CCPREC = "-DDOUBLE_PREC";


$PREC = "single";               # precision
$elesize = 4;                   # element size in bytes
$min_blk = int($reg_num/8);     

### DEFINE ###
$max_size = 300;	        # maximum size of the test square matrix
                                # will be changed by L1
$inner_standard = 1000000;	# try to do 1MFLOP in the inner loop
$outer_standard = 10;		# and outer loop

$m_start = 1;                   # starting values for block size
$n_start = 1;
$k_start = 1;
@normal_transpose = ("NN", "NT", "TN");	
                                # all the cases of normal/transpose


### MAIN PROGRAM ###

# Take in the arguments from the command line

&parse_args;
open (STDOUT, "| tee $LOG_FILE");

print "Testing started at ", `date`;
if (!$start_at){		# start from register level
    &gen_sizes;

    if (defined($matfile)){
	&open_matfile;
    }
}

# This is what we need to do no matter what level to start
# Compile the matrix-matrix generation program

&make_gen;

# Generating different block sizes with differnt compiler 
# option based on the total number of the machine registers

`rm -f *.o *~ core`;

# Compile rprf calling routine once with appropriate precision

print "Compiling $PERFNAME......\n";
if ($PREC eq "single"){
    `$COMPILER $CCFLAGS $CCDEF -c $PERFNAME.c`;
} else {
    `$COMPILER $CCFLAGS $CCDEF $CCPREC -c $PERFNAME.c`;
}

die ("Cannot create $PERFNAME.o\n") if ($?);

# Register level search for all the normal_transpose matrices

if (!$start_at){
    for ($i=0; $i<@normal_transpose; $i++){
	$RES_FILE = "./mm_res_L0_${normal_transpose[$i]}.mat";
	open(OUTFILE,">$RES_FILE")||die("Can't open $RES_FILE for output\n");
	&print_header;
	foreach $option (@options){
	    &gen_blk($option, $normal_transpose[$i]);
	}
	close (OUTFILE);
    }
}

# L1 level search

if ($depth > 1 && $start_at <= 1){	# proceed to L1 search if dep > 1
    $ver = 0;                         # ver of the result file

    for ($i=0; $i<@normal_transpose; $i++){
	$RES_FILE = "./mm_res_L${ver}_${normal_transpose[$i]}.mat";
	$best_blk = &select_best("peak"); # get the best block size <peak>

	if ($best_blk == -1) {
	    printf ("No $RES_FILE, skip.\n");
	} else {
	    print "best blk from file $RES_FILE :$best_blk";
    
	    $next_ver = $ver+1;
	    $RES_FILE =~ s/L$ver/L$next_ver/g;  # changed the output filename
	    print "Working on L${next_ver} blocking......\n";
	    &first_level($best_blk, $normal_transpose[$i]);
	}
    }
    --$depth;
}


# L2 level search

if ($depth > 1 && $start_at <= 2){  # proceed to L2 search if depth still > 1

    # We need different approach here, because on L1, there are 4 cases
    # of normal/transpose, so we need to do loop here

    $ver = 1;

    for ($i=0; $i<@normal_transpose; $i++){
        # get the best block size <mean> for a specific case

	# set the file name and get the block
	$RES_FILE = "./mm_res_L${ver}_$normal_transpose[$i].mat";
	$best_blk = &select_best("mean"); 
	if ($best_blk == -1){
	    printf( "No $RES_FILE, skip.\n");
	} else {

	    print "best blk from file $RES_FILE:  $best_blk";

	    # increment the next level file name and search
	    $next_ver = $ver+1;
	    $RES_FILE =~ s/L$ver/L$next_ver/;   # changed the output filename
	    print"Working on L$next_ver $normal_transpose[$i] blocking...\n";
	    &second_level($best_blk, $normal_transpose[$i]);
	}
    }

    --$depth;			# for future L3 search
}

&select_nbest;
print "Testing completed at ", `date`;
# Clean up the mess
`rm -f $FLNAME.c $FLNAME.o $PERFNAME.o $TARGET`;
close(STDOUT);
exit (0);

### SUBROUTINES ###

# Subroutine for generating different depth of optimization

sub gen_blk {

    local ($option, $norm_trans) = @_;
    local ($m, $k, $n);
    local ($mflops, $prev_mflops);
    local ($count);


  M_LOOP:
    for ($m=$m_start; $m<=$reg_num; $m++){
      N_LOOP:
	for ($n=$n_start; $n<=$reg_num; $n++){
	  K_LOOP:
	    for ($k=$k_start; $k<=$k_limit; $k++){

		# some criteria to limit our search space
		last N_LOOP if (($m*$n) > $reg_num);
		last K_LOOP if (($m*$n) < $min_blk);
		
		# Perform the test, always NN in reg block level
		$mflops = &kernel ($option, $norm_trans, $m, $k, $n);
		next K_LOOP if ($mflops == -1);

		# determine if we should proceed for the next k
		if ($mflops < $prev_mflops){
		    if ($count >= $k_allow){
			$prev_mflops = 0;    # reset
			$count = 0;
			last K_LOOP;
		    } else {
			$count++;
			$prev_mflops = $mflops;    # update
		    }
		} else {
		    $count = 0;
		    $prev_mflops = $mflops;
		}
	    }
	    $prev_mflops = 0;            # reset
	    $count = 0;
	}
    }
}


# Subroutine for carrying out the testing procedure

sub kernel {

    local ($option, $norm_trans, @flags) = @_;
    local ($tmp, @tmp, %tmp, $i);
    local (@foo, $foo);
    local ($mflops, $total, $max);   
    
    # generate the mm-code
    $retval = &gen_code($norm_trans, @flags);
    die ("Something wrong for @flags\n") if ($retval);

    # compile the newly generated routine 

    print ("Compiling $FLNAME.c with @flags, $option:\n");
    $retval = &compile($option);
		    
    if (!$retval){       # compiled successfully
			
	# Run the program with different sizes
	print"Testing : @flags with $option";
	print OUTFILE "@flags, $option: ";

	# Initialize the statistics
	$total = 0;
	@foo = ();
	%tmp = @tmp = ();

	# run the performance test
	# doesn't matter for the code efficiency, because
	# the majority of the runtime isn't from the script,
	# but from the compile and benchmark :-P

	if (!defined($matfile)){ # default, dense matrix testing only

	    foreach $size (@sizes){
		$foo = &run_perf_test($size, 0); 
		$foo =~ s/^\s+|\s+\n$//g;
		if ($foo){
		    $total += 1/$foo;
		    push(@foo, $foo);
		} else {
		    print OUTFILE "error ";
		}
		print ".";
	    }
	    print "\n";
	} elsif (!$append) { 	# problems in file only
	    foreach $size (@problems){
		$foo = &run_perf_test($size, 1); 
		$foo =~ s/^\s+|\s+\n$//g;
		if ($foo){
		    $total += 1/$foo;
		    push(@foo, $foo);
		} else {
		    print OUTFILE "error ";
		}
		print ".";
	    }
	    print "\n";
	} else {		# test both dense matrix && customize problems

	    foreach $size (@sizes){ # dense matrix first
		$foo = &run_perf_test($size, 0); 
		$foo =~ s/^\s+|\s+\n$//g;
		if ($foo){
		    $total += 1/$foo;
		    push(@foo, $foo);
		} else {
		    print OUTFILE "error ";
		}
		print ".";
	    }

	    foreach $size (@problems){ # then the customized problems
		$foo = &run_perf_test($size, 1); 
		$foo =~ s/^\s+|\s+\n$//g;
		if ($foo){
		    $total += 1/$foo;
		    push(@foo, $foo);
		} else {
		    print OUTFILE "error ";
		}
		print ".";
	    }
	    print "\n";
	}
	
	# Performance analysis
	for ($i=0; $i<@foo; $i++){
	    # It was the proposed filtering method, but we
	    # decided to drop the filtering for now after
	    # running some statistics experiments
	    # $tmp = (0.5*$foo[$i-1]+$foo[$i] +0.5*$foo[$i+1])/2;
	    $tmp = $foo[$i];
	    push (@tmp, $tmp);
	    push (@tmp, $sizes[$i]);
	}
	%tmp = @tmp;
	@tmp = sort sort_num(keys(%tmp));
	$max = pop(@tmp);

	$mflops = @sizes / $total;
	printf OUTFILE("; %5.3f; @%d %5.3f\n",$mflops, $tmp{$max}, $max);
    } else {
	warn ("Something wrong when compiling the routine\n");
	return -1;
    }
    $mflops;
}

# subroutine for calling mm-generation program

sub gen_code {
    
    local($norm_trans, @flags) = @_;
    local($i);
    local(@cmd);
    local($poa, $pob);		# single chcaracter to hold opA opB

    if (@flags % 3){	# something wrong if it's not multiple of 3
	warn ("Wrong number of arguments.\n");
	return 1;
    }
    # Use chop instead of substr for fun
    $pob = chop($norm_trans);	# b first
    $poa = chop($norm_trans);	# then a

    @cmd = ("./".$GEN_TARGET);
    while (@flags){
	push (@cmd, $CBFLAG);
	for($i=0; $i<3; $i++){
	    push (@cmd, shift(@flags));
	}
    }
    splice(@cmd, $#cmd+1,0,$ROUTINEFLAG,$ROUTINE_NAME,
	   $PRECFLAG, $PREC, $FLFLAG,$FLNAME.".c", 
	   $OPA, $poa, $OPB, $pob);

    system(@cmd);
    $retcode = $? >> 8;
}


# subroutine for compiling the performance program

sub compile {

    local(@cmd);
    local($opt) = @_;

    `$COMPILER $opt $CCFLAGS $CCDEF -c $FLNAME.c`;

    if (!$?){
	`$COMPILER $opt $CCFLAGS -o $TARGET $PERFNAME.o $FLNAME.o $CCLIBS`;
    }
    $?;
}


# Subroutine for running the performance test
# This invoke the performance test program, and write the 
# max value for each matrix size
# 
# arguments: $size - either the size of the test matrix, or the string
#                    of problem configuration, depends on $whatproblem
#            $whatproblem - 0 to operate on dense matrix test
#                           1 to operate on general matrix size

sub run_perf_test {
    
    local($size, $whatproblem) = @_; # whatproblem: 0 when it's default
    local($result, @result);	     # 1 when it's a problem from file
    local(@cmd, $inner, $outer);
    local($name) = "./" . $TARGET;
    local ($op);
    local (@problem, $m, $k, $n, $rowA, $colA, $rowB, $colB, $rowC, $colC);

    if (!$whatproblem){		# Dense matrix, default
	$op = $size * $size * (2*$size-1);
    } else {
	@problem = split(' ', $size); # size is the problem configuration
	$m = shift(@problem);
	$k = shift(@problem);
	$n = shift(@problem);
	$rowA = shift(@problem);
	$colA = shift(@problem);
	$rowB = shift(@problem);
	$colB = shift(@problem);
	$rowC = shift(@problem);
	$colC = shift(@problem);
	$op = $m*$n*(2*$k-1);
    }

    # at least run xMFLOPS specified in inner_standard or 1 time
    # if the op is > xMFLOPS
    if ($op < $inner_standard){
	$inner = "-i ". int($inner_standard/$op);
    } else {
	$inner = "-i 1";
    }
    $outer = "-o ". $outer_standard;

    if (!$whatproblem){		# Dense matrix, default
	if ($size > 500) {	# allocate more memory if size's getting big
	    @cmd = ($name, "-A 26", $inner, $outer, $size, $size, $size,
		    $size, $size, $size, $size, $size, $size);
	} else {
	    @cmd = ($name, $inner, $outer, $size, $size, $size,
		    $size, $size, $size, $size, $size, $size);
	}

    } else {
	if ($m > 500){
	    @cmd = ($name, "-A 26", $inner, $outer, $m, $k, $n,
		    $rowA, $colA, $rowB, $colB, $rowC, $colC);
	} else {
	    @cmd = ($name, $inner, $outer, $m, $k, $n,
		    $rowA, $colA, $rowB, $colB, $rowC, $colC);
	}
    }
    open(DATA, "@cmd | awk '{print \$10}'|");
    
    $result = <DATA>;
    $result =~ s/\n$//g;
    print OUTFILE "$result ";
    $result;
    
}

# Subroutine for launching the first level cache search

sub first_level {

    local ($best_blk, $norm_trans) = @_;
    local (@best_blk, @tmp);               # $best_blk[0] = blk sizes
                                           # $best_blk[1] = option
                                           # $best_blk[2] = peak location

    local ($bestm, $bestn, $bestk); 
    local (@var_m, @var_n, @var_k);        # array for m,n, k variation
    local ($i);

    # Take out the best parameters from the previous level

    @best_blk = split (/;|,/, $best_blk);  # ',' as delimiter btw size & option
    $best_blk[0] =~ s/^\s+|\s+$//g;
    $best_blk[1] =~ s/^\s+|\s+$//g;
    $best_blk[2] =~ s/\s+@(\d+)\b.+\n$/$1/;
    @tmp = split(/ /, $best_blk[0]);       # ' ' as delimiter for blk size

    # Compute the new limits of the test

    $bestm = $best_blk[2]/$tmp[$#tmp-2];   # best values from previous level
    $bestk = $best_blk[2]/$tmp[$#tmp-1];
    $bestn = $best_blk[2]/$tmp[$#tmp];

    if (!$L2){                             # new matrix size limit
	$max_size = 300;
    }
    if ($max_mat1){
	$max_size = $max_mat1;
    }
    if ($L2 && !$max_mat1){
	$max_size = int (sqrt(2*$L2*1024/(3*$elesize)));  
    }

    for ($i=0; $i<@PEAK_ALLOW; $i++){
	push(@var_m, int($bestm * $PEAK_ALLOW[$i]));
	push(@var_n, int($bestn * $PEAK_ALLOWN[$i]));
	push(@var_k, int($bestk * $PEAK_ALLOW[$i]));
    }

    &gen_sizes;
    if (defined($matfile)){
	&open_matfile;
    }


    open(OUTFILE,">$RES_FILE")||die("Can't open $RES_FILE for output\n");
    &print_header;

    # looping different L1 blocking combination
    foreach $m (@var_m){
	foreach $n (@var_n){
	    foreach $k (@var_k){
		next if (!($m) || !($n) || !($k));
		&kernel ($best_blk[1], $norm_trans, @tmp, $m, $k, $n);
	    }
	}
    }
    close (OUTFILE);

}

# Subroutine for launching the second level cache search

sub second_level {

    local ($best_blk, $norm_trans) = @_;
    local (@best_blk, @tmp);               # $best_blk[0] = blk sizes
                                           # $best_blk[1] = option
                                           # $best_blk[2] = peak location

    local ($bestm, $bestn, $bestk); 
    local (@var_m, @var_n, @var_k);        # array for m,n, k variation
    local ($i);

    # Take out the best parameters from the previous level

    @best_blk = split (/;|,/, $best_blk);  # ',' as delimiter btw size & option
    $best_blk[0] =~ s/^\s+|\s+$//g;
    $best_blk[1] =~ s/^\s+|\s+$//g;
    $best_blk[2] =~ s/\s+@(\d+)\b.+\n$/$1/;
    @tmp = split(/ /, $best_blk[0]);       # ' ' as delimiter for blk size

    # values of trials using the L2 cache size constraint
    $bestm = $bestk = $bestn = int(sqrt(($L2*1024)/($elesize*3)));

    if (!$max_mat2){		# new matrix size limit
	$max_size = 1000;	# at least 1000
    }
    if ($max_mat2){
	$max_size = $max_mat2;
    }

    for ($i=0; $i<@PEAK_ALLOW; $i++){
	push(@var_m, int($bestm * $PEAK_ALLOW[$i]));
	push(@var_n, int($bestn * $PEAK_ALLOWN[$i]));
	push(@var_k, int($bestk * $PEAK_ALLOW[$i]));
    }

    &gen_sizes;

    if (defined($matfile)){
	&open_matfile;
    }

    open(OUTFILE,">$RES_FILE")||die("Can't open $RES_FILE for output\n");
    &print_header;

    foreach $m (@var_m){
	foreach $n (@var_n){
	    foreach $k (@var_k){
		next if (!($m) || !($n) || !($k));
		&kernel ($best_blk[1], $norm_trans, @tmp, $m, $k, $n);

	    }
	}
    }

    close (OUTFILE);
}



# Subroutine for selecting the maximum performance from the register blocking

sub select_best{

    local ($criterion) = @_;
    local ($count) = 1;
    local ($data, @data, %res_array);
    local (@orderkey);
    local ($max, $limit);           #variables relate to best perf and min k
    local ($k, $mink, $perf, $minkperf);

    if (!-e $RES_FILE) {              # skip rather than kill the process
	return -1;
    }
    open (RESFILE, $RES_FILE) || die ("Cannot open $RES_FILE\n");

    $data = <RESFILE>;     # junk
    $data = <RESFILE>;

    while ($data = <RESFILE>){
	if ($criterion eq "peak"){
	    $data =~ s/:.*;/;/g;     # ';' as delimiter for mean && max
	    @data = split(/ /, $data);
	    $res_array{$data[$#data]} = $data;
	} elsif ($criterion eq "mean"){
	    @data = split(/,|:|;/, $data);
	    $res_array{$data[$#data-1]} = 
		$data[0].", ".$data[1]."; ".$data[4];
	}
    }
    
    close (RESFILE);
    @orderkey = sort sort_num(keys(%res_array));

    # get the best performance && compute the acceptable variation
    $max = $minkperf = $perf = pop(@orderkey);
    $limit = $max * (1-$var);
    $mink = $k_limit;
    
    while ($limit < $perf){
	$data = $res_array{$perf};
	@data = split(' ', $data);
	$k = $data[1];
	if ($k < $mink){
	    $minkperf = $perf;
	    $mink = $k;
	}
	$perf = pop(@orderkey);
    }

    return $res_array{$minkperf};
}

# Subroutine for selecting n-best from the harmonic mean

sub select_nbest {

    local ($data, @data, %res_array);
    local ($mean, $para, $peak);        # variables for mean && parameters
    local ($i);
    local ($level, $count, @tp);
    local (@all_nt) = ("NN", "NT", "TN", "TT");

    open (OUTFILE, ">$outfile") || die ("Cannot open $outfile\n");

    print OUTFILE "### Architecture :", `$GUESS`, "\n";
    print STDOUT "### Architecture :", `$GUESS`, "\n";

    # set the appropriate file
    for ($level=0; $level <= $next_ver; $level++){
	for ($i=0;$i<@all_nt; $i++){
	    $RES_FILE = "./mm_res_L${level}_$all_nt[$i].mat";

	    if (-e $RES_FILE) {
		open (RESFILE, $RES_FILE) || die ("Cannot open $RES_FILE\n");

		print STDOUT  "### level $level : $all_nt[$i] ###\n";
		print OUTFILE "### level $level : $all_nt[$i] ###\n";

		# Print testing parameter

		$data = <RESFILE>;     # junk
		print STDOUT  "# ", $data; # only print once
		print OUTFILE "# ", $data; # only print once

		$data = <RESFILE>;
		print STDOUT  "# ", $data;
		print OUTFILE "# ", $data;
		
		%res_array = (); # initialize the array
		$count = 1;	 # and count variable

		if ($level){	# L1 or L2 search
		    
		    while ($data = <RESFILE>){
			@data = split(/,|:|;/, $data);
			$mean = $data[$#data-1];
			$data[$#data] =~ s/\n//g;
			$para = $data[0].", ".$data[1].": ".$data[$#data];
			# blk + opt + peak
			$res_array{$mean} = $para;
		    }
		} else {	# registor search
		    while ($data = <RESFILE>){
			@data = split(/,|:|;/, $data);
			$peak = $data[$#data];
			@tp = split(' ', $peak);
			$peak = $tp[$#tp];
			$para = $data[0].", ".$data[1].": ".$data[$#data-1].
			    "; ".$tp[$#tp-1];
			# blk + opt + mean + peak location
			@data = split(/ /, $data);
			$res_array{$peak} = $para;
		    }
		}
		    
		
		foreach $blk (reverse sort sort_num (keys(%res_array))){
		    if ($count <= $nbest){
			print STDOUT  "$res_array{$blk} $blk\n";
			print OUTFILE "$res_array{$blk} $blk\n";
			$count++;
		    } else {
			last;
		    }
		}
		close (RESFILE);
	    } else {
		print STDOUT  "# $RES_FILE doesn't exist, skipped.\n";
		print OUTFILE "# $RES_FILE doesn't exist, skipped.\n";
	    }
	}
    }
    close (OUTFILE);
}


# Subroutine for parsing the arguments

sub parse_args {

    local ($i, $j, $opt);
    local ($gotminblk)=0;	# check if minblk option is specified
    local ($gotapp) = 0;	# check if append && matfile happen at
				# the same time
    local (@tmp);

    if (!(@ARGV)){
	&help_menu;
	exit(1);
    }

    for ($i=0; $i<@ARGV; $i++){
	if ($ARGV[$i] =~ /^-./){

	    # Test for compiler option
	    if ($ARGV[$i] =~ /^-cc\b/){
		$COMPILER = $ARGV[++$i];
	    }
	    
	    # Test for the memory hierarchy depth
	    elsif ($ARGV[$i] =~ /^-dep\b/){
		$depth = $ARGV[++$i];
		die ("Memory hierarchy must be an integer.\n") 
		    if ($depth =~ /\D+/);
		die ("Current version only support L2 blocking.\n")
		    if ($depth > 3);
	    }

	    # Test for the compiler option file
	    elsif ($ARGV[$i] =~ /^-coptfile\b/){
		$opt = $ARGV[++$i];
		$opt =~ s/\s+//g;
		die ("$opt doesn't exist.\n") if (!(-e $opt));
		open (OPFILE, $opt) || die ("Cannot open $opt.\n");
		@options = <OPFILE>;
		close (OPFILE);
		chop(@options);
	    }

	    # Test for the nbest option
	    elsif ($ARGV[$i] =~ /^-nbest\b/){
		$nbest = $ARGV[++$i];
		die ("Must be an integer.\n") if ($nbest =~ /\D+/);
	    }

	    # Test for the maximum k option
	    elsif ($ARGV[$i] =~ /^-maxk\b/){
		$k_limit = $ARGV[++$i];
		die ("Max k must be an integer.\n") if ($k_limit =~ /\D+/);
	    }

	    # Test for the consective decrease option
	    elsif ($ARGV[$i] =~ /^-cond\b/){
		$k_allow = $ARGV[++$i];
		die ("Consective decrease limit must be an integer.\n") 
		    if ($k_allow =~ /\D+/);
	    }

	    # Test for the register number of the machine
	    elsif ($ARGV[$i] =~ /^-reg\b/){
		$reg_num = $ARGV[++$i];
		die ("Must be an integer for register number.\n")
		    if ($reg_num =~ /\D+/);
	    }

	    # Test for the L2 cache size
	    elsif ($ARGV[$i] =~ /^-L2\b/){
		$L2 = $ARGV[++$i];
		die ("Must be an integer for L2 cache size.\n")
		    if ($L2 =~ /\D+/);
	    }

	    # Test for the L1 cache size
	    elsif ($ARGV[$i] =~/^-L1\b/){
		$L1 = $ARGV[++$i];
		die ("Must be an integer for L1 cache size.\n")
		    if ($L1 =~ /\D+/);
	    }

	    # Test for precision to use
	    elsif ($ARGV[$i] =~ /^-prec\b/){
		$PREC = $ARGV[++$i];
		die ("Precision must be either single or double\n")
		    if ($PREC ne "single" && $PREC ne "double");
		if ($PREC eq "double"){
		    $elesize = 8;
		}
	    }

	    # Test for maximum matrix to use for L1
	    elsif ($ARGV[$i] =~ /^-maxsize1\b/){
		$max_mat1 = $ARGV[++$i];
		die("Must be an integer for maximum matrix size.\n")
		    if ($max_mat1 =~/\D+/);
	    }

	    # Test for maximum matrix to use for L2
	    elsif ($ARGV[$i] =~ /^-maxsize2\b/){
		$max_mat2 = $ARGV[++$i];
		die ("Must be an integer for maximum matrix size.\n")
		    if ($max_mat2 =~ /\D+/);
	    }

	    # Test for minimum blocking combination
	    elsif ($ARGV[$i] =~ /^-minblk\b/){
		$min_blk = $ARGV[++$i];
		die ("Must be an interger.\n")
		    if ($min_blk =~ /\D+/);
		$gotminblk = 1;
	    }
	    
	    # Test for the variant option
	    elsif ($ARGV[$i] =~ /^-var\b/){
		$var = $ARGV[++$i];
		die ("Acceptable limit must be an integer.\n")
		    if ($var =~ /\D+/);
		$var /= 100;
	    }

	    # Test for the matrix size file
	    elsif ($ARGV[$i] =~ /^-pconfig\b/){
		die ("pconfig and apconfig can't coexist\n")
		    if ($gotapp);
		$matfile = $ARGV[++$i];
		$append = 0;
		$gotapp = 1;
	    }

	    # Test for append matrix size file
	    elsif ($ARGV[$i] =~ /^-apconfig\b/){
		die ("apconfig and pconfig can't coexist\n")
		    if ($gotapp);
		$matfile = $ARGV[++$i];
		$append = 1;
	    }

	    # Test for the start level
	    elsif ($ARGV[$i] =~ /^-start\b/){
		$start_at = $ARGV[++$i];
		die ("Must be an interger.\n") if ($start_at =~ /\D+/);
	    }

	    # Test for the output file
	    elsif ($ARGV[$i] =~ /^-outfile\b/){
		$outfile = $ARGV[++$i];
	    }

	    # Test for the normal and transpose
	    elsif ($ARGV[$i] =~ /^-matop\b/){
		# clear up the normal/transpose array if this option on
		@normal_transpose = ();
		while (!($ARGV[++$i] =~ /^-./) && $i < @ARGV){
		    # check the validity
		    die ("No such opertaion on matrix $ARGV[$i]\n")
			if ($ARGV[$i] ne "NN" &&
			    $ARGV[$i] ne "TN" &&
			    $ARGV[$i] ne "NT" &&
			    $ARGV[$i] ne "TT");
		    push (@tmp, $ARGV[$i]);
		}
		$i--;
		die ("Exceed the maximum number of combination.\n")
		    if (@tmp > 4);
		die ("There must be at least one matrix operation.\n")
		    if (!@tmp);
		# direct map to the normal transpose array
		for ($j=0; $j<@tmp; $j++){
		    $normal_transpose[0] = "NN" if ($tmp[$j] eq "NN");
		    $normal_transpose[1] = "NT" if ($tmp[$j] eq "NT");
		    $normal_transpose[2] = "TN" if ($tmp[$j] eq "TN");
		    $normal_transpose[3] = "TT" if ($tmp[$j] eq "TT");
		    }
		# delete the null elements
		for ($j=0; $j<@normal_transpose; $j++){
		    if ($normal_transpose[$j] eq ""){
			splice(@normal_transpose, $j--, 1);
		    }
		}
	    }

	    # Test for the help option
	    elsif ($ARGV[$i] =~ /^-help\b/){
		&help_menu;
		exit(0);
	    }

	    else {
		print "Unknown option $ARGV[$i], ignored.\n";
	    }
	} else {
	    print "Unknown argument $ARGV[$i], ignored\n";
	}
    }

    # put a space there if there isn't any option
    if (!@options){
	push(@options, " ");
    }

    # adjust the paramenters entered by the command prompt
    $max_size = int(sqrt(2*$L1*1024/(3*$elesize)));
    if (!$gotminblk){
	$min_blk = int($reg_num/8);
    }
    die ("min blocking is greater than register number.\n")
	if ($min_blk > $reg_num);

    # check for L2 value if the search depth is 3
    die ("L2 value must be supplied if the search depth is 3.\n")
	if ($depth > 2 && !$L2);

    # check for the start level and the depth
    die ("The start level cannot be greater than the memory hierarchy.\n")
	if ($depth - $start_at <= 0);

}

# Subroutine for printing out the help prompt

sub help_menu {

    print "Usage: $0 [OPTIONS]\n";
    print "where [OPTIONS] include:\n";
    print "Machine descriptions\n";
    print "  -reg g<32> \t\t\t: number of registers in the machine\n";
    print "  -L1 size<8> \t\t\t: L1 cache size in KB\n";
    print "  -L2 size \t\t\t: L2 cache size in KB\n";
    print "Routine description\n";
    print "  -prec [<single>|double]\t: Precision to use\n";
    print "  -matop operations \t\t: matrix op. any subset of NN NT TN TT\n";
    print "Compiler options\n";
    print "  -cc name<cc>\t\t: an ANSI C compiler \n";
    print "  -coptfile name \t\t: filename for compiler option\n";
    print "Search parameters\n";
    print " Memory hierarchy\n";
    print "  -dep dep<1> \t\t\t: depth of memory hierarchy\n";
    print "  -start level<0>\t\t: start the search at level\n";
    print "  -maxsize1<300> \t\t: maximum matrix size to be tested for L1\n";
    print "  -maxsize2<1000>\t\t: maximum matrix size to be tested for L2\n";
    print "  -maxk k<4> \t\t\t: maximum searching space\n";
    print "  -minblk l\t\t\t: minimum blocking combination to be tested\n";
    print " Workload\n";
    print "  -pconfig name \t\t: Use matrix file as the problem config\n";
    print "  -apconfig name \t\t: Use matrix file in addition to default\n";
    print " Search heuristic\n";
    print "  -var v<1%>\t\t\t: acceptable limit of the peak performance\n";
    print "  -cond d<5> \t\t\t: maximum allow for consective decrease\n";
    print " Search result reporting\n";
    print "  -outfile filename \t\t: output file name\n";
    print "  -nbest n<10>\t\t\t: n-th number of best blocking\n";
    print "  -help \t\t\t: Print this message\n";
}

# Subroutine for printing out header on the output file

sub print_header {

    print OUTFILE "command: $0 @ARGV\n";

    print OUTFILE "@sizes\n";

    select (OUTFILE);		# disable output buffer
    $| = 1;
    select (STDOUT);
    $| = 1;
    select (STDERR);
    $| = 1;
    select (STDOUT);
}



# Subroutine for making mm-generation program 

sub make_gen {

    if (-e $GEN_TARGET){
	print "Removing $GEN_TARGET......\n";
	`rm -f $GEN_TARGET`;
    }

    print "Making $GEN_TARGET......\n";
    `$COMPILER -O -o $GEN_TARGET $GEN_TARGET.c`;
}



# Generate different size of a square matrix 

sub gen_sizes {

    @sizes = ();
    local ($cutoff);

    if ($sizecutoff > $max_size){
	$cutoff = $max_size;
    } else {
	$cutoff = $sizecutoff;
    }

    # Getting the size of power of 2
    for ($i=2; $i<=$max_size; $i*=2){
	push (@sizes, $i);
    }
    
    # Getting the size of power of 3
    for ($i=3; $i<=$max_size; $i*=3){
	push (@sizes, $i);
    }
    
    # Getting the size of multiples of 10
    for ($i=10; $i<=$cutoff; $i+=10){
	push (@sizes, $i);
    }

    # Getting the size of mutiples of 100
    for ($i=100; $i<=$max_size; $i+=100){
	push (@sizes, $i);
    }

    # powers of 1.5 starting at 13 for primes
    $i = 12;
    while ($i <= $max_size){
	$i = int($i)+1;
	push (@sizes, $i);
	$i = $i ** 1.5;
    }
    @sizes = sort sort_num (@sizes);
    # delete the repeat size
    for ($i=0;$i<@sizes-1;$i++){
	if ($sizes[$i] == $sizes[$i+1]){
	    splice(@sizes, $i--, 1);
	}
    }
}

# Subroutine to read the matrix file

sub open_matfile {
    
    local ($data);
    local (@tmp, $m, $k, $n, $rowA, $colA, $rowB, $colB, $rowC, $colC);
    local ($count) = 0;

    if (!$append) {		# only use the customized test case
	@sizes = ();		# by clearing up the internal sizes
    }
    open (MATFILE, "$matfile") || die ("Cannot open $matfile.\n");
    while ($data = <MATFILE>){
	$data =~ s/\n$//g;	# strip the newline char
	next if ($data =~ /^\s*\#/); # ignore comment
	next if (!length($data)); 
	# check to see if they are correct
       
	@tmp = split(' ', $data);
	die ("Invalid number of arguments in $matfile.\n") if (@tmp != 9);
	$m = shift(@tmp);
	$k = shift(@tmp);
	$n = shift(@tmp);
	$rowA = shift(@tmp);
	$colA = shift(@tmp);
	$rowB = shift(@tmp);
	$colB = shift(@tmp);
	$rowC = shift(@tmp);
	$colC = shift(@tmp);

	die ("Invalid matrix A in line $data\n") if ($rowA < $m || $colA < $k);
	die ("Invalid matrix B in line $data\n") if ($rowB < $k || $colB < $n);
	die ("Invalid matrix C in line $data\n") if ($rowC < $m || $colC < $n);
	# pass the 9 parameters in one big chunk to make life easier
	$problems[$count++] = $data;

	# Put them into variable sizes
	push(@sizes, $m);	# assumed square matrix
    }
}		 


# Subroutine to change the sorting behavior

sub sort_num {			# sort numerically
    $a <=> $b;
}
