/*
 *   Performance measurement for mul_mfmf_mf routine.
 *   The code is based on a rprf interface rather the ipm interface
 *
 *   Originally written by Jeff Bilmes
 *   Modified by Dominic LAM <ctlam@po.eecs.berkeley.edu>
 *
 *   $Id: mul_mm_m_perf.c,v 1.1 1995/10/10 05:56:35 ctlam Exp $
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include "rprf.h"

#ifdef DOUBLE_PREC
#define MM_FP_TYPE double
#else
#define MM_FP_TYPE float
#endif 


extern void
php_mul_mfmf_mf(int,int,int,MM_FP_TYPE*,MM_FP_TYPE*,MM_FP_TYPE*,int,int,int);
void copy_d_vd(size_t, double, double*);
void copy_f_vd(size_t, float, float*);

/* Simple matrix initialize routine */

/* double precision */
void copy_d_vd(size_t len, double from, double *to)
{
  size_t i;
  
  for (i=0; i<len; i++)
    *to++ = from;
}

/* single precision */
void copy_f_vf(size_t len, float from, float *to)
{
  size_t i;
  
  for (i=0; i<len; i++)
    *to++ = from;
}


int
main(int argc, const char *argv[])
{

    MM_FP_TYPE *A, *B, *cC;
    size_t M=10, K=10, N=10;
    size_t rowA=10, colA=10;	/* size of the matrices */
    size_t rowB=10, colB=10;
    size_t rowC=10, colC=10; 

    if (argc < 9) {
      printf ("Usage: %s [<perf_args>] M K N rowA colA rowB colB rowC colC \n"
	      ,argv[0]);
      exit(0);
    }
  
    /* Get the size of the matrix */
    rprf_parse_args(argc, argv, "iiiiiiiii", &M, &K, &N, 
		    &rowA, &colA, &rowB, &colB, &rowC, &colC);

    /* Error checking */

    if (rowA < M || colA < K){
      printf ( "Invalid matrix size for matrix A\n");
      exit(1);
    } 
    if (rowB < K || colB < N){
      printf ( "Invalid matrix size for matrix B\n");
      exit(2);
    }
    if (rowC < M || colC < N){
      printf ( "Invalid matrix size for matrix C\n");
      exit(3);
    }

#ifdef DOUBLE_PREC
    /* double precision */
    
    /* Allocate and initialize. */
    
    A = rprf_malloc_vd(rowA*colA);
    B = rprf_malloc_vd(rowB*colB);
    cC = rprf_malloc_vd(rowC*colC);

    copy_d_vd(rowA*colA, 1.0, A);
    copy_d_vd(rowB*colB, 1.0, B);

#else

    /* Single precision */    

    /* Allocate and initialize. */
    
    A = rprf_malloc_vf(rowA*colA);
    B = rprf_malloc_vf(rowB*colB);
    cC = rprf_malloc_vf(rowC*colC);

    copy_f_vf(rowA*colA, 1.0f, A);
    copy_f_vf(rowB*colB, 1.0f, B);


#endif                /* End of DOUBLE */
    

/* Run the performance test */

#define FUNC php_mul_mfmf_mf(M,K,N,A,B,cC,colA,colB,colC);

    RPRF_BODY(FUNC, M*N*(2*K-1))

    return 0;
}
