#! /usr/local/bin/perl
#
# nbest - this script select the n-best blocking size combination
#         using either harmonic mean or peak performance as criterion
#         from the raw data file produced by perf_script.pl
#
# Written by Dominic LAM <ctlam@po.eecs.berkeley.edu>
#
# $Header: /n/icsie/db/real/ctlam/final_script/RCS/nbest.pl,v 1.2 1995/10/10 02:16:19 ctlam Exp ctlam $ 


$RES_FILE = "";
$COND = "mean";
$nbest = 10;
$count = 1;
$target = -1;          # matrix size to be search (dafault = -1)

&parse_args;

$data = <RESFILE>;     # junk
$data = <RESFILE>;

# Extract the matrix sizes from the data file

if ($target >0){

    @sizes = split(' ', $data);

    $position = -1;              # position of the target matrix size
    for ($i=0; $i<@sizes; $i++){
	if ($target == $sizes[$i]){
	    $position = $i;
	}
    }
}

die ("There is no matrix size $target in the data file.\n")
    if ($position < 0);

# Reading the data

while ($data = <RESFILE>){

    @data = split(/,|:|;/, $data);

    # $data[0] = blk combo
    # $data[1] = compiler option
    # $data[2] = raw data
    # $data[3] = harmonic mean
    # $data[4] = max of raw data

    $para = $data[0].", ".$data[1].": ";


    ### Put the data in the array according to the flags
    ### either 1) perf at size S
    ###        2) peak
    ###        3) harmonic mean

    if ($target >= 0){                     # extract the performance
	@rawdata = split(' ', $data[2]);
	$perf_at_target = $rawdata[$position];
	$res_array{$perf_at_target} = $para;
    }                          
    elsif ($COND eq "peak"){
	$peak = $data[$#data];             # extract the peak, 
	@tmp = split(' ', $peak);
	$peak = $tmp[$#tmp];
	$para = $para . " $tmp[$#tmp-1]";  # peak location
	$res_array{$peak} = $para;
    } elsif ($COND eq "mean"){
	$mean = $data[$#data-1];           # mean
	$res_array{$mean} = $para;
    }
}

foreach $blk (reverse sort sort_num (keys(%res_array))){
    if ($count <= $nbest){
	print "$res_array{$blk}\t $blk\n";
	$count++;
    } else {
	last;
    }
}
close (RESFILE);
exit(0);


### Subroutine for parsing the arguments

sub parse_args {
    
    local ($i);
    local ($hadcond, $hadsize) = 0;
    
    if (@ARGV < 1){
	&help_menu;
	exit(1);
    }

    $RES_FILE = $ARGV[0];
    open(RESFILE, $RES_FILE) || die ("Cannot open $RES_FILE\n");

    for ($i=1; $i<@ARGV; $i++){
	if ($ARGV[$i] =~ /^-./){

	    # Test for the n-best option
	    if ($ARGV[$i] =~ /^-n\b/){
		$nbest = $ARGV[++$i];
		die ("Invalid n\n")
		    if ($nbest =~ /\D+/);
	    }

	    # Test for the criterion
	    elsif ($ARGV[$i] =~ /^-cond\b/){
		$COND = $ARGV[++$i];
		$hadcond = 1;             # has a -cond flag
		die ("Either 'mean' or 'peak' can be selected.\n")
		    if ($COND ne "peak" && $COND ne "mean");
		die ("-cond cannot coesits with -size\n")
		    if ($hadsize == 1);
	    }

	    # Test for the size search option
	    elsif ($ARGV[$i] =~ /^-size\b/){
		$target = $ARGV[++$i];
		$hadsize = 1;
		die ("Matrix size must be an integer.\n") 
		    if ($target =~ /\D+/);
		die ("-size cannot coexist with -cond\n")
		    if ($hadcond == 1);
	    }

	    # Test for help option
	    elsif ($ARGV[$i] =~ /^-help\b/){
		&help_menu;
		exit(0);
	    }

	    # defaults
	    else {
		print "Unknown option $ARGV[$i], ignored.\n";
	    }
	} else {
	    print "Unknown argument $ARGV[$i], ignored.\n";
	}
    }
}

sub help_menu {

    print "Usage : $0 filename [OPTIONS]\n";
    print "where [OPTIONS] include:\n";
    print " -help \t\t\t: Print this message\n";
    print " -n nbest\t\t: first nbest blocking parameters\n";
    print " -cond peak|<mean>\t: rank the performance by peak or mean\n";
    print " -size s \t\t: rank the performance at size s\n";
}



sub sort_num {			# sort numberically
    $a <=> $b;
}
  
