/*
  $Header: /u/spert/src/rprf/RCS/rprf.h,v 1.7 1995/09/12 23:50:34 krste Exp $

  Routines to help with performance testing.
*/

#ifndef rprf_h_
#define rprf_h_

#ifdef __cplusplus
extern "C" {
#endif

#include "IPM_timer.h"

extern long rprf_n_inner;
extern long rprf_n_outer;

extern void
rprf_parse_args(int argc, const char* argv[], const char* format, ...);

/* Functions to allocate vectors from memory arena. */
#define rprf_malloc_vi8(len) \
    ((char *) rprf_malloc(len, sizeof(char)))

#define rprf_malloc_vi16(len) \
    ((short *) rprf_malloc((len)*sizeof(short), sizeof(short)))

#define rprf_malloc_vi32(len) \
    ((long *) rprf_malloc((len)*sizeof(long), sizeof(long)))

#define rprf_malloc_vf(len) \
    ((float *) rprf_malloc((len)*sizeof(float), sizeof(float)))
        
#define rprf_malloc_vd(len) \
    ((double *) rprf_malloc((len)*sizeof(double), sizeof(double)))
        
extern void *rprf_malloc(size_t len, size_t alignment);

#define RPRF_BODY(one_run, ops_arg) \
{ \
    double rprf_ops = (double) (ops_arg); \
    IPM_timer rprf_t; \
\
    long rprf_i; \
    long rprf_j; \
    double rprf_mops; \
    double best_mops = 0.0; \
\
    for (rprf_i=0; rprf_i<rprf_n_outer; rprf_i++) \
    { \
        const long rprf_n = rprf_n_inner; \
        IPM_timer_clear(&rprf_t); \
\
        IPM_timer_start(&rprf_t); \
        for (rprf_j=0; rprf_j<rprf_n; rprf_j++) \
        { \
            one_run; \
        } \
        IPM_timer_stop(&rprf_t); \
\
        rprf_mops = (1e-6 * (rprf_ops) * rprf_n)/IPM_timer_read(&rprf_t); \
        if (rprf_mops > best_mops) \
            best_mops = rprf_mops; \
    } \
    printf("%8.3f\n", best_mops); \
}

#ifdef __cplusplus
}
#endif

#endif
